
from tkinter import W
import numpy as np
import matplotlib.pyplot as plt
from itertools import count
from collections import namedtuple

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.distributions import Categorical

from env2 import *

from log import *

from model22 import *

env = Grid()

learning_rate = 0.01
gamma = 0.99
episodes = 20000

render = True
eps = np.finfo(np.float32).eps.item()

SavedAction = namedtuple('SavedAction', ['log_prob', 'value'])


n = 5

LOG = 1

class Policy(nn.Module):
    def __init__(self):
        super(Policy, self).__init__()
        self.fc1 = nn.Linear(2*n, 32)
        #debug(self.fc1.weight)
        self.actor_embedding = ResEncoder(10, 2, 128, 16, 512, 3)
        #self.critic_embedding = ResEncoder(10, 2, 128, 16, 512, 3)

        self.actor_decoder = ResDecoder(128, 360)

        '''
        self.actoru = nn.Sequential(
            nn.Linear(32, n)
        )
        self.actorw = nn.Sequential(
            nn.Linear(32, 2*n)
        )
        '''


        self.critic_embedding = ResEncoder(10, 2, 128, 16, 512, 3)
        self.g = nn.Linear(128, 1, bias=False)
        self.critic = nn.Sequential(
            nn.Linear(128, 256, bias=True),
            nn.ReLU(),
            nn.Linear(256, 1, bias=True)
        )

        self.save_actions = []
        self.rewards = []

    def get_actor_embedding(self, x):
        #debug("embedding: ", x)
        #x = self.fc1(x)
        #debug("embedding: ", self.fc1.weight)
        #x = F.relu(x)
        #action_score = self.actoru(x)
        #state_value = self.critic(x)
        #debug("embedding: ", x)
        am = torch.zeros(n, n)
        kpm = torch.zeros(1, n)

        actor_emb = self.actor_embedding(x, am, kpm)

        self.actor_emb = actor_emb

        return actor_emb
    
    def get_critic(self, x):
        am = torch.zeros(n, n)
        kpm = torch.zeros(1, n)
        #debug(self.critic_embedding)

        critic_emb = self.critic_embedding(x, am, kpm)

        debug(critic_emb.shape)

        glimpse = self.g(torch.tanh(critic_emb)).view(1, n)
        debug(glimpse.shape)
        glimpse = torch.softmax(glimpse, dim=1).unsqueeze(1)
        glimpse = torch.bmm(glimpse, critic_emb)

        debug(glimpse.shape)


        x = self.critic(glimpse)

        return x

    def getu(self, visited):
        #x = nn.ReLU(self.fc1(x))
        #debug("get u: ", x)
        #x = self.actoru(x)
        
        #debug("get u: ", x)
        #debug("get u: ", x)
        
        x = self.actor_decoder.getu(self.actor_emb, visited)
        u_prob = torch.where(visited, -1e6 * torch.ones_like(x), x)
        return F.softmax(u_prob, dim=-1)

    def getw(self, u, unvisited):
        #x = nn.ReLU(self.fc1(x))
        #x = self.actorw(x)  
        debug("getting w")      
        debug(u)
        x = self.actor_decoder.getw(self.actor_emb, u, unvisited)
        debug(x.shape, x)
        debug(unvisited)
        w_prob = torch.where(unvisited, -1e6 * torch.ones_like(x), x)
        return F.softmax(w_prob, dim=-1)

    def update_decoder(self, u, w=-1, v=-1, h=-1):
        self.actor_decoder.update(self.actor_emb, u, w, v, h)

       

        
model = Policy()
optimizer = optim.Adam(model.parameters(), lr=learning_rate)
for i in model.parameters():
    debug(i.shape)

def select_action(state):

    #debug(state)
    actions = []

    points = torch.tensor(state[0])/10
    points = points.view(1, n, 2)
    #points = torch.flatten(points)

    visited = torch.tensor(state[1]).to(torch.bool)
    visited = torch.logical_not(torch.tensor([True, False, False, False, False]))

    debug(points)
    #debug(visited, unvisited)

    points = torch.tensor(points).float()
    model.actor_decoder.reset_state()
    for i in range(n-1):
        #embedding = model(points)
        #model.get_actor_embedding(points)
        model.actor_emb = torch.rand([1, 5, 128])


        #select u
        #debug(points)
        probs = model.getu(visited)
        p = Categorical(probs)
        u = p.sample()
        #debug("u: ", u, probs, visited)
        #debug()

        visited[u] = False
        #debug("u: ", u)

        #select w        
        unvisited = torch.logical_not(visited)
        unvisited = torch.cat((unvisited, unvisited), dim=0)
        probs = model.getw(u, unvisited)
        p = Categorical(probs)

        w0 = p.sample()
        #debug("w: ", w, probs, unvisited)
        #debug()
        


        if w0 < n:
            v = u
            h = w0
            w = w0
        else:
            w = w0-n
            v = w
            h = u

        visited[w] = False

        if i < n-2:
            model.update_decoder(u, w, v, h)

        actions.append([u, w, v, h])
    
    debug(points)
    state_value = model.get_critic(points)

    state_value = state_value.view(1)
    debug(state_value)
    
    model.save_actions.append(SavedAction(p.log_prob(w0), state_value))
    
    #debug(actions)
    return actions






    
def finish_episode():
    R = 0
    save_actions = model.save_actions
    policy_loss = []
    value_loss = []
    rewards = []

    for r in model.rewards[::-1]:
        R = r + gamma * R
        rewards.insert(0, R)

    rewards = torch.tensor(rewards)
    #rewards = (rewards - rewards.mean()) / (rewards.std() + eps)

    debug()
    debug("training")
    debug(save_actions)
    debug(rewards)

    for (log_prob , value), r in zip(save_actions, rewards):
        reward = r - value.view(-1).item()
        policy_loss.append(-log_prob * reward)
        value_loss.append(F.smooth_l1_loss(value, torch.tensor([r])))
    
    #debug(policy_loss)
    #debug(value_loss)
    optimizer.zero_grad()
    loss = torch.stack(policy_loss).sum() + torch.stack(value_loss).sum()
    #debug(loss)
    loss.backward(retain_graph=True)
    optimizer.step()
    #optimizer.zero_grad()
    #debug("policy_loss", policy_loss)
    #loss = policy_loss[0]
    #loss.backward(retain_graph=True)
    #optimizer.step()


    del model.rewards[:]
    del model.save_actions[:]

def main():

    for i_episode in count(episodes):
        state = env.reset()
        debug(state)

        actions = select_action(state)
        debug("actions: ", actions)
        state, reward, done= env.multistep(actions)
        model.rewards.append(reward)
        #env.render()
        finish_episode()

if __name__ == '__main__':
    main()