from email.header import Header
from re import L
import torch
import random
import numpy as np

import os

from log import *


HEIGHT, WIDTH = [8,8]

OBSTACLE = 1000

BLANK=0
LINE=1
NODE=2
OBSTACLE=3


class Interval:
    def __init__(self, low, high):
        self.low = min(low,high)
        self.high = max(low, high)

    def intersect(self, inter):
        return max(self.low, inter.low) < min(self.high, inter.high)

    def merge(self, inter):
        assert self.intersect(inter)

        self.low = min(self.low, inter.low)
        self.high = max(self.high, inter.high)

def log(a, k=0):
    print("\033[3", k, "m", a, "\033[0m", end="", sep="")

class Point:
    def __init__(self, x=-1, y=-1):
        self.x = x
        self.y = y


class Grid:
    def __init__(self, h=10, w=10, n=5):
        self.h = h
        self.w = w
        self.n = n
        self.grid = [[BLANK]*self.w for i in range(self.h)]
        self.hlines = [[]] * h
        self.wlines = [[]] * w

        self.steps = 0

        random.seed(1)
        #self.generate_nodes()

    def reset(self):
        self.steps = 0
        self.grid = [[BLANK]*self.w for i in range(self.h)]
        self.hlines = [[]] * self.h
        self.wlines = [[]] * self.w

        self.generate_nodes()

        state = [[[p.x, p.y] for p in self.nodes], self.nodes_used]
        #print(state)
        return state

    def generate_nodes(self):
        random.seed(1)
        self.hashnodes = set()
        while len(self.hashnodes) < self.n:
            x = random.randint(0, self.h*self.w - 1)
            self.hashnodes.add(x)
            #print(x)

        self.hashnodes = [32, 97, 72, 17, 8]

        self.nodes = []
        for i in self.hashnodes:
            self.nodes.append(Point(i // self.w, i % self.w))
            #print(self.nodes[-1].x, self.nodes[-1].y)
            self.grid[self.nodes[-1].x][self.nodes[-1].y] = NODE

        self.nodes_used = [0] * self.n

    def occupy(self, x, y):
        if self.grid[x][y] == BLANK:
            self.grid[x][y] = LINE            

    def connect(self, start, end, dir):
        #assert self.nodes_used[start] == 1
        #assert self.nodes_used[end] == 0

        self.nodes_used[start] = 1
        self.nodes_used[end] = 1
        n0 = self.nodes[start]
        n1 = self.nodes[end]

        inter0 = Interval(n0.x, n1.x)
        inter1 = Interval(n0.y, n1.y)

        if dir:
            #mid = Point(n0.x, n1.y)

            self.hlines[n0.y].append(inter0) ######################### merge #############
            self.wlines[n1.x].append(inter1)

            for i in range(inter0.low, inter0.high+1, 1):
                self.occupy(i, n0.y)
            for i in range(inter1.low, inter1.high+1, 1):
                self.occupy(n1.x, i) 
        else:
            
            self.hlines[n1.y].append(inter0) ######################### merge #############
            self.wlines[n0.x].append(inter1)

            for i in range(inter0.low, inter0.high+1, 1):
                self.occupy(i, n1.y)
            for i in range(inter1.low, inter1.high+1, 1):
                self.occupy(n0.x, i) 

    def wire_length(self):
        res = 0
        for i in range(self.h):
            for j in range(self.w):
                res += (self.grid[i][j] != BLANK)
        return res-1


    def finish(self):
        return self.sum(self.nodes_used) == self.n


    def step(self, start, end, dir):
        self.connect(start, end, dir)
        if self.finish():
            reward = -self.wire_length()
            done = 1
        else:
            reward = 0
            done = 0
        
        state = [[[p.x, p.y] for p in self.nodes], self.nodes_used]
        return state, reward, done

    
    def multistep(self, actions):
        for i in actions:
            dir = int(i[0] == i[1])
            self.connect(i[0], i[1], dir)

        reward= -self.wire_length()
        
        state = [[[p.x, p.y] for p in self.nodes], self.nodes_used]
        return state, reward, 1

    def render(self):
        #os.system("cls")
        print()

        for i in range(self.h):
            log(str(self.h-i-1)+"   ", 4)
            for j in range(self.w):
                log("o ", self.grid[self.h-i-1][j])
            print()
        
        print()
        log("    ")
        for i in range(self.w):
            log(str(i)+" ", 4)
        print()

        for i in range(self.n):
            print(i, "(", self.nodes[i].x, self.nodes[i].y, "): ", self.nodes_used[i])
        

    def finish(self):
        #assert sum(self.nodes_used) == self.steps+1

        return self.steps+1 == self.n

    def step(self, start, end, dir):
        self.steps += 1
        self.connect(start, end, dir)
        if self.finish():
            reward = -self.wire_length()
            done = 1
        else:
            reward = 0
            done = 0
        
        return np.array(self.nodes_used), reward, done, None

if __name__ == "__main__":
    env = Grid()
    print(env.reset())
    #_, reward, done = env.step(0, 1, 1)
    #_, reward, done = env.step(1, 2, 0)
    #_, reward, done = env.step(2, 3, 0)
    #_, reward, done = env.step(3, 4, 0)
    #print(reward, done)

    actions = [[0, 1, 1, 0],
            [1, 2, 1, 2],
            [2, 3, 2, 3],
            [3, 4, 3, 4],
    ]
    _, reward, done = env.multistep(actions)
    env.render()