from collections import namedtuple
import tkinter as tk
import torch
import torch.nn as nn

from log import *

class ResEncoder(nn.Module):
    def __init__(self, max_n, gdim=2, edim=128, nheads=16, hdim=512, num_layer=3):
        super(ResEncoder, self).__init__()
        self.max_n = max_n
        self.edim = edim
        self.nheads = nheads
        self.hdim = hdim
        self.num_layer = num_layer

        self.fc0 = nn.Linear(gdim, edim)


        #self.key_padding_mask = torch.zeros(batch_size, 50)
        #self.attn_mask = torch.zeros(50, 50)
        
        if torch.__version__[0:3] == "1.8":
            self.encoder_layer = nn.TransformerEncoderLayer(d_model=edim, nhead=nheads, dim_feedforward=hdim, dropout=0)#norm_first, batch_first=True)
        else:
            self.encoder_layer = nn.TransformerEncoderLayer(d_model=edim, nhead=nheads, dim_feedforward=hdim, dropout=0, norm_first=True, batch_first=True)
        
        self.encoder = nn.TransformerEncoder(self.encoder_layer, num_layer)

    def forward(self, x, attn_mask, key_padding_mask):
        x = self.fc0(x)
        #print(x)
        #attn_mask = torch.zeros(self.max_n, self.max_n)
        #key_padding_mask = torch.zeros(10, self.50)

        x = self.encoder(x, mask=attn_mask, src_key_padding_mask=key_padding_mask)
        return x


class PTM(nn.Module):
    def __init__(self, edim, qdim):
        super(PTM, self).__init__()
        self.edim = qdim
        self.qdim = qdim

        self.efc = nn.Linear(edim, qdim, bias=False)
        self.qfc = nn.Linear(qdim, qdim, bias=False)
        self.th0 = nn.Tanh()
        self.gfc = nn.Linear(qdim, 1, bias=False)
        self.th1 = nn.Tanh()

    def forward(self, e, q, visited=None):
        #print(e.shape, q.shape)
        e = self.efc(e)
        q = self.qfc(q)
        e = self.gfc(self.th0(e+q))
        #print(e.shape)
        #print("visited: ", visited.shape)
        e = torch.squeeze(e)
        #if visited != None:
        #    l = torch.where(visited, -1e6 * torch.ones_like(e), e)
        #print(l.shape)
        l = e
        l = self.th1(l) * 10.0
        #l = torch.softmax(l, dim=1)
        return l

State = namedtuple("State", ["edge", "subtree", "eu", "ew", "ev", "eh"])
        
class ResDecoder(nn.Module):

    def __init__(self, edim, qdim):
        super(ResDecoder, self).__init__()
        self.edim = edim
        self.qdim = qdim
        self.ptm = PTM(edim, qdim)

        self.eptm0 = PTM(edim, qdim)
        self.eptm1 = PTM(edim, qdim)

        Edge = namedtuple("Edge", ["ufc", "wfc", "vfc", "hfc"])
        self.edgefc = Edge(nn.Linear(edim, qdim),
                         nn.Linear(edim, qdim),
                         nn.Linear(edim, qdim),
                         nn.Linear(edim, qdim),
                        )
        self.subtfc = nn.Linear(qdim, qdim)

        self.fc_w5 = nn.Linear(edim, qdim)

        self.state = []
        self.state.append(State(
            torch.zeros(self.qdim),
            torch.zeros(self.qdim),
            -1,
            -1,
            -1,
            -1
        ))

        #self.subtree = torch.zeros(self.qdim)

    def reset_state(self):
        self.state = []
        self.state.append(State(
            torch.zeros(self.qdim),
            torch.zeros(self.qdim),
            -1,
            -1,
            -1,
            -1
        ))

    def getu(self, e, visited, step=0):
        
        qu = torch.max(self.state[-1].edge + self.state[-1].subtree, torch.zeros_like(self.state[-1].edge))
        #debug(qu)
        u = self.ptm(e, qu, visited)

        return u

    def getw(self, e, u, visited, step=1):

        qw = torch.max(torch.zeros_like(self.state[-1].edge), self.state[-1].edge + self.state[-1].subtree + self.fc_w5(e[:, self.state[-1].eu, :]))

        w0 = self.eptm0(e, qw, visited)
        w1 = self.eptm1(e, qw, visited)
        w = torch.cat([w0, w1]).view(1, -1)
        
        return w 

    def update(self, e, u, w=-1, v=-1, h=-1):
        
        edg = self.edgefc.ufc(e[:, u, :].clone())

        subtr = torch.max(self.state[-1].subtree, self.subtfc(edg))

        self.state.append(State(edge=edg, subtree=subtr, eu=u, ew=w, ev=v, eh=h))

        return self.state[-1]


        
        

        



    
        
            




if __name__ == "__main__":    


    ### encoder test
    print("encoder test")
    
    a = torch.zeros([10, 50, 2])    
    if torch.__version__[0:3] == "1.8":
        a = torch.zeros([50, 10, 2]) 
        a=a
    print(a.shape)
    RE = ResEncoder(10, 2, 128, 16, 512, 3)
    amask = torch.zeros(50, 50)
    kpmask = torch.zeros(10, 50)
    b = RE(a, amask, kpmask)
    print(b.shape)

    '''
    a = torch.ones(2, 2)
    b = torch.ones(2, 4)
    ptm = PTM(2, 4)
    c = ptm(a, b, torch.tensor([0, 1, 0, 1])>0)
    print(c)
    '''

    ### decoder test     

    a = torch.zeros([1, 50, 2])
    RE = ResEncoder(10, 2, 16, 16, 16, 1)

    RD = ResDecoder(16, 32)
    
    amask = torch.zeros(50, 50)
    kpmask = torch.zeros(1, 50)
    embedding = RE(a, amask, kpmask)
    embedding = torch.squeeze(embedding)
    print("embedding: ", embedding.shape)

    optimizer = torch.optim.Adam(RD.parameters()+RE.parameters(), lr=0.1)

    ### decoder step 0
    print("decoder step 0: ")
    from torch.distributions import Categorical
    visited = torch.zeros([50]).to(torch.bool)
    u_prob = RD.getu(embedding, visited)
    #print(u_prob.shape)
    u_prob = torch.softmax(u_prob, dim=0)
    #print(u_prob.shape)
    catu = Categorical(u_prob)
    u = catu.sample()
    print(u)

    
    state = RD.update(embedding, u)
    print(state)

    ### decoder step 1
    print("decoder step 1: ")
    u_prob = RD.getu(embedding, visited)
    u_prob = torch.softmax(u_prob, dim=0)
    catu = Categorical(u_prob)
    u = catu.sample()
    print(u)

    w_prob = RD.getw(embedding, u, visited)
    w_prob = torch.softmax(w_prob, dim=0)
    print(w_prob.shape)
    
    catw = Categorical(w_prob)
    w = catw.sample()
    print(w)
    if w < 50:
        v = u
        h = w
    else:
        w = w-50
        v = w
        h = u


    state = RD.update(embedding, u, w, v, h)
    print(state)



    ### decoder step 2
    print("decoder step 2: ")
    u_prob = RD.getu(embedding, visited)
    u_prob = torch.softmax(u_prob, dim=0)
    catu = Categorical(u_prob)
    u = catu.sample()
    print(u)

    w_prob = RD.getw(embedding, u, visited)
    w_prob = torch.softmax(w_prob, dim=0)
    print(w_prob.shape)
    
    catw = Categorical(w_prob)
    w = catw.sample()
    print(w)
    if w < 50:
        v = u
        h = w
    else:
        w = w-50
        v = w
        h = u


    state = RD.update(embedding, u, w, v, h)
    print(state)

    reward= -30

    loss = -u_prob[u] * reward
    reward.backward()
    