import torch
import pdb

from agent import Actor, Critic, Policy
from env.env import create_RSMTEnv

def update_mask(mask, batch_size, node_visited):
  for i in range(batch_size):
    mask[i][node_visited[i]] = True
  return mask

batch_size = 4
num_nodes = 8
nodes = torch.randn(batch_size, num_nodes, 2)

env = create_RSMTEnv(num_nodes=num_nodes, pos_l=0, pos_h=100)

actor_net = Actor()
critic_net = Critic()
model = Policy(actor=actor_net, 
               critic=critic_net, 
               obs_space=env.observation_space,
               action_space=env.action_space
              )

u0 = model.first_act(nodes)
v_list = [u0]
h_list = []

mask_visited = torch.zeros(batch_size, num_nodes).bool()

mask_visited = update_mask(mask_visited, batch_size, u0)

for i in range(1, num_nodes):
  v, h = model.act(nodes, mask_visited)
  mask_visited = update_mask(mask_visited, batch_size, v)
  mask_visited = update_mask(mask_visited, batch_size, h)
  v_list.append(v)
  h_list.append(h)

# transpose into [#num_batch, #num_nodes]
all_v = torch.stack(v_list).transpose(1, 0)
all_h = torch.stack(h_list).transpose(1, 0)

rewards = -10 * torch.ones(batch_size)
model.learn(nodes, rewards)

pdb.set_trace()